/*
 * Name:    example2.c
 *
 * Purpose: This program demonstrates running a multiple frame sequence.
 *          The program has one command line argument 'use_client_disk'
 *          which if set to a non-zero value will make the client program
 *          transfer the frames to its local disk and save them there.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2006 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_image.h"
#include "mx_area_detector.h"

static mx_status_type
save_frame_on_detector_computer( MX_RECORD *ad_record,
                                long last_frame_number );

static mx_status_type
save_frame_on_client_computer( MX_RECORD *ad_record,
                                MX_IMAGE_FRAME **image_frame,
                                long last_frame_number );

int
main( int argc, char *argv[] )
{
        MX_RECORD *mx_database, *ad_record;
        MX_IMAGE_FRAME *image_frame;
        char mx_database_name[] = "./mx_client.dat";
        char area_detector_name[] = "ad";
        double exposure_time, gap_time;
        long num_frames, last_frame_number, total_num_frames;
        long current_frame_number;
        unsigned long ad_status_flags;
        int use_client_disk;
        mx_status_type mx_status;

        if ( argc != 2 ) {
                fprintf(stderr, "Usage: %s 'use_client_disk_flag'\n", argv[0]);
                exit(1);
        }

        image_frame = NULL;

        /* Do we save the frames on the client computer or do we save
         * the frames on the detector computer?
         */

        use_client_disk = atoi( argv[1] );

        /* Initialize the MX database. */

        mx_status = mx_setup_database( &mx_database, mx_database_name );

        if ( mx_status.code != MXE_SUCCESS ) {
                fprintf( stderr,
                        "Did not successfully initialize MX database '%s'.\n",
                        mx_database_name );
                exit(1);
        }

        /* Find the area detector record in the database. */

        ad_record = mx_get_record( mx_database, area_detector_name );

        if ( ad_record == NULL ) {
                fprintf( stderr,
                "Did not find area detector record '%s' in MX database '%s'.\n",
                        area_detector_name, mx_database_name );
                exit(1);
        }

        /* Configure a sequence for the area detector that tells it to
         * take 10 image frames with an exposure time of 1.0 second
         * for each frame and a gap of 0.5 seconds between the frames.
         */

        num_frames    = 10;
        exposure_time = 1.0;
        gap_time      = 0.5;

        fprintf( stderr,
"Taking a multiframe sequence of %ld frames using area detector '%s' with "
"an exposure time of %g seconds and a gap time of %g seconds.\n",
                num_frames, ad_record->name,
                exposure_time, gap_time );

        mx_status = mx_area_detector_set_multiframe_mode( ad_record,
                                                        num_frames,
                                                        exposure_time,
                                                        gap_time );

        if ( mx_status.code != MXE_SUCCESS )
                exit(1);

        /* Arm the detector. */

        fprintf( stderr, "Arming the detector.\n" );

        mx_status = mx_area_detector_arm( ad_record );

        if ( mx_status.code != MXE_SUCCESS )
                exit(1);

        /* We assume for this example that the area detector is configured
         * for internal trigger mode.  If so, then we will need to explicitly
         * trigger the data acquisition sequence.  If we were using an
         * external trigger, then this step would not be necessary.
         */

        fprintf( stderr, "Triggering the detector.\n" );

        mx_status = mx_area_detector_trigger( ad_record );

        if ( mx_status.code != MXE_SUCCESS )
                exit(1);

        /* Monitor the progress of the data acquisition sequence. */

        current_frame_number = -1;

        while (1) {
                /* Ask for the status of the detector. */

                mx_status = mx_area_detector_get_extended_status( ad_record,
                                                        &last_frame_number,
                                                        &total_num_frames,
                                                        &ad_status_flags );

                if ( mx_status.code != MXE_SUCCESS )
                        exit(1);

                /* If the reported last frame number has changed, then there
                 * are one or more frames available to be read out.
                 */

                if ( last_frame_number != current_frame_number ) {

                        current_frame_number++;

                        if ( use_client_disk ) {
                                mx_status = save_frame_on_client_computer(
                                        ad_record, &image_frame,
                                        current_frame_number );
                        } else {
                                mx_status = save_frame_on_detector_computer(
                                        ad_record, current_frame_number );
                        }

                        if ( mx_status.code != MXE_SUCCESS )
                                exit(1);
                }

                /* If the status flags say that the area detector is
                 * no longer busy, see if there are any frames left
                 * to read out.
                 */

                if ( ( ad_status_flags & MXSF_AD_IS_BUSY ) == 0 ) {

                        if ( current_frame_number >= last_frame_number ) {

                            /* If not, then exit. */

                            fprintf( stderr,
                            "The data acquisition sequence has completed.\n" );
                            fprintf( stderr, "Exiting now...\n" );

                            exit(0);
                        }
                }

                /* Sleep for a millisecond so that we do not use up
                 * all of the CPU time on the computer.
                 */

                mx_msleep(1);
        }

        fprintf( stderr, "Should never get here.\n" );

        exit(1);
}

static mx_status_type
save_frame_on_detector_computer( MX_RECORD *ad_record,
                                long last_frame_number )
{
        char savefile_name[80];
        mx_status_type mx_status;

        fprintf( stderr,
    "Saving area detector '%s' frame number %ld on the detector computer.\n",
                ad_record->name, last_frame_number );

        /* Tell the detector computer to readout the requested frame from
         * the area detector hardware into the detector computer's primary
         * image frame buffer.
         */

        mx_status = mx_area_detector_readout_frame( ad_record,
                                                last_frame_number );

        if ( mx_status.code != MXE_SUCCESS )
                return mx_status;

        /* Tell the detector computer to correct the frame that was
         * just read out.
         */

        mx_status = mx_area_detector_correct_frame( ad_record );

        if ( mx_status.code != MXE_SUCCESS )
                return mx_status;

        /* Tell the detector computer to save the frame on its own
         * hard disk.
         *
         * If we specify a full pathname, the MX server will save
         * the file at the requested location.
         *
         * If we give the server only the filename, it will save the
         * file in the default save file directory on the detector
         * computer.
         *
         * If we give it an empty or NULL filename, the MX server
         * will choose the next filename on its own.
         */

        snprintf( savefile_name, sizeof(savefile_name),
                "example2_%04ld.pnm", last_frame_number );

        mx_status = mx_area_detector_save_frame( ad_record,
                                                MXFT_AD_IMAGE_FRAME,
                                                savefile_name );

        if ( mx_status.code != MXE_SUCCESS )
                return mx_status;

        fprintf(stderr, "Successfully wrote image file '%s'.\n", savefile_name);

        return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
save_frame_on_client_computer( MX_RECORD *ad_record,
                                MX_IMAGE_FRAME **image_frame,
                                long last_frame_number )
{
        MX_IMAGE_FRAME *local_image_frame;
        char savefile_name[80];
        size_t image_length;
        void *image_data_pointer;
        uint16_t *uint16_array;
        int i;
        mx_status_type mx_status;

        fprintf( stderr,
    "Saving area detector '%s' frame number %ld on the client computer.\n",
                ad_record->name, last_frame_number );

        /* The first time that mx_area_detector_setup_frame() is invoked,
         * it will allocate memory for the image frame data structures.
         * On subsequent calls, it checks to see if the already allocated
         * image frame data structures are too small to hold the new
         * image frame.  If they are already big enough, the image frame
         * is left alone.
         *
         * In addition, mx_area_detector_setup_frame() saves a pointer
         * to the image frame in the area detector record data structure.
         * Later in this routine, mx_area_detector_transfer_frame() will
         * read the image sent by the detector computer into that frame.
         */

        mx_status = mx_area_detector_setup_frame( ad_record, image_frame );

        if ( mx_status.code != MXE_SUCCESS )
                return mx_status;

        /* Tell the detector computer to readout the requested frame from
         * the area detector hardware into the detector computer's primary
         * image frame buffer.
         */

        mx_status = mx_area_detector_readout_frame( ad_record,
                                                last_frame_number );

        if ( mx_status.code != MXE_SUCCESS )
                return mx_status;

        /* Tell the detector computer to correct the frame that was
         * just read out.
         */

        mx_status = mx_area_detector_correct_frame( ad_record );

        if ( mx_status.code != MXE_SUCCESS )
                return mx_status;

        /* Tell the detector computer to send the primary image frame
         * across the network to the client computer.  If this call
         * completes successfully, the image will be accessible via
         * the *image_frame pointer.
         */

        local_image_frame = *image_frame;

        mx_status = mx_area_detector_transfer_frame( ad_record,
                                                MXFT_AD_IMAGE_FRAME,
                                                local_image_frame );

        if ( mx_status.code != MXE_SUCCESS )
                return mx_status;

        /* Display the first 10 pixels in the frame.  You should always
         * reinvoke mx_image_get_image_data_pointer() after each time
         * you read in a frame.
         */

        mx_status = mx_image_get_image_data_pointer( local_image_frame,
                                                        &image_length,
                                                        &image_data_pointer );

        if ( mx_status.code != MXE_SUCCESS )
                return mx_status;

        fprintf( stderr, "The transferred image is %lu bytes long.\n",
                        (unsigned long) image_length );

        uint16_array = image_data_pointer;

        for ( i = 0; i < 10; i++ ) {
                fprintf( stderr, "image[%d] = %hu\n",
                        i, (unsigned short) uint16_array[i] );
        }

        /* We finish by writing the local image to a disk file in PNM format.
         * 
         * FIXME: For the real release, we need to convert this to SMV format.
         */

        snprintf( savefile_name, sizeof(savefile_name),
                "example2_%04ld.pnm", last_frame_number );

        mx_status = mx_image_write_file( local_image_frame,
                                        MXT_IMAGE_FILE_PNM,
                                        savefile_name );

        if ( mx_status.code != MXE_SUCCESS )
                return mx_status;

        fprintf(stderr, "Successfully wrote image file '%s'.\n", savefile_name);

        return MX_SUCCESSFUL_RESULT;
}

