/*
 * Name:    example4.c
 *
 * Purpose: This program acquires a single image frame, reads out a region
 *          of interest and then writes the ROI to a disk file on the client.
 *
 * Note:    If programmed in this manner, only the pixels contained within
 *          the ROI will be transferred across the network.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2006 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_image.h"
#include "mx_area_detector.h"

int
main( int argc, char *argv[] )
{
        MX_RECORD *mx_database, *ad_record;
        MX_IMAGE_FRAME *roi_frame;
        void *roi_data_pointer;
        uint16_t *uint16_array;
        char mx_database_name[] = "./mx_client.dat";
        char area_detector_name[] = "ad";
        unsigned long i, roi_number, roi[4];
        size_t roi_length;
        double exposure_time;
        char roi_filename[MXU_FILENAME_LENGTH+1];
        mx_bool_type busy;
        mx_status_type mx_status;

        /* Initialize the MX database. */

        mx_status = mx_setup_database( &mx_database, mx_database_name );

        if ( mx_status.code != MXE_SUCCESS ) {
                fprintf( stderr,
                        "Did not successfully initialize MX database '%s'.\n",
                        mx_database_name );
                exit(1);
        }

        /* Find the area detector record in the database. */

        ad_record = mx_get_record( mx_database, area_detector_name );

        if ( ad_record == NULL ) {
                fprintf( stderr,
                "Did not find area detector record '%s' in MX database '%s'.\n",
                        area_detector_name, mx_database_name );
                exit(1);
        }

        /* Configure the detector to acquire a single frame. */

        exposure_time = 0.5;        /* in seconds */

        fprintf( stderr,
        "Taking a single frame exposure of %g seconds.\n", exposure_time );

        mx_status = mx_area_detector_set_one_shot_mode( ad_record,
                                                        exposure_time );

        if ( mx_status.code != MXE_SUCCESS )
                exit(1);

        /* Arm the detector. */

        fprintf( stderr, "Arming the detector.\n" );

        mx_status = mx_area_detector_arm( ad_record );

        if ( mx_status.code != MXE_SUCCESS )
                exit(1);

        /* We assume for this example that the area detector is configured
         * for internal trigger mode.  If so, then we will need to explicitly
         * trigger the data acquisition sequence.  If we were using an
         * external trigger, then this step would not be necessary.
         */

        fprintf( stderr, "Triggering the detector.\n" );

        mx_status = mx_area_detector_trigger( ad_record );

        if ( mx_status.code != MXE_SUCCESS )
                exit(1);

        /* Monitor the progress of the data acquisition sequence. */

        while (1) {
                /* Check to see if the detector is still acquiring a frame. */

                mx_status = mx_area_detector_is_busy( ad_record, &busy );

                if ( mx_status.code != MXE_SUCCESS )
                        exit(1);

                if ( busy == FALSE ) {

                        /* If the area detector has finished the measurement,
                         * then break out of the while() loop.
                         */
                        
                        break;
                }

                /* Sleep for a millisecond so that we do not use up
                 * all of the CPU time on the computer.
                 */

                mx_msleep(1);
        }

        fprintf( stderr, "Exposure complete.\nReading out frame 0.\n" );

        /* Tell the detector computer to readout the requested frame from
         * the area detector hardware into the detector computer's primary
         * image frame buffer.
         */

        mx_status = mx_area_detector_readout_frame( ad_record, 0 );

        if ( mx_status.code != MXE_SUCCESS )
                exit(1);

        /* Tell the detector computer to correct the frame that was
         * just read out.
         */

        fprintf( stderr, "Correcting the frame.\n" );

        mx_status = mx_area_detector_correct_frame( ad_record );

        if ( mx_status.code != MXE_SUCCESS )
                exit(1);

        /* Define the boundaries of region of interest 5.  This can be done
         * either before or after acquiring the image frame.
         */

        roi_number = 5;

        /* The following limits are specified in binned coordinates. */

        roi[0] = 1000;    /* X minimum */
        roi[1] = 2000;    /* X maximum */
        roi[2] = 250;     /* Y minimum */
        roi[3] = 750;     /* Y maximum */

        fprintf( stderr,
        "Setting ROI %lu to Xmin = %lu, Xmax = %lu, Ymin = %lu, Ymax = %lu\n",
                roi_number, roi[0], roi[1], roi[2], roi[3] );

        mx_status = mx_area_detector_set_roi( ad_record, roi_number, roi );

        if ( mx_status.code != MXE_SUCCESS )
                exit(1);

        /* Make sure that mx_area_detector_get_roi_frame() allocates
         * a new frame by setting roi_frame to NULL.
         */

        roi_frame = NULL;

        /* Transfer the contents of the detector ROI to the client ROI frame. */

        fprintf( stderr, "Reading out ROI %lu\n", roi_number );

        mx_status = mx_area_detector_get_roi_frame( ad_record, NULL,
                                                roi_number, &roi_frame );

        if ( mx_status.code != MXE_SUCCESS )
                exit(1);

        /* Display the first 10 pixels in the ROI frame. */

        mx_status = mx_image_get_image_data_pointer( roi_frame,
                                                &roi_length,
                                                &roi_data_pointer );

        if ( mx_status.code != MXE_SUCCESS )
                exit(1);

        fprintf( stderr, "The transferred ROI frame is %lu bytes long.\n",
                        (unsigned long) roi_length );

        uint16_array = roi_data_pointer;

        for ( i = 0; i < 10; i++ ) {
                fprintf( stderr, "image[%lu] = %hu\n",
                        i, (unsigned short) uint16_array[i] );
        }

        /* We finish by writing the ROI frame to a disk file in PNM format.
         * 
         * FIXME: For the real release, we need to convert this to SMV format.
         */

        strlcpy( roi_filename, "roifile.pgm", MXU_FILENAME_LENGTH );

        mx_status = mx_image_write_file( roi_frame,
                                        MXT_IMAGE_FILE_PNM,
                                        roi_filename );

        if ( mx_status.code != MXE_SUCCESS )
                exit(1);

        fprintf(stderr, "Successfully wrote ROI file '%s'.\n", roi_filename );

        exit(0);
}

